@@ -1,4 +1,4 @@
-/* g10.c - The GNUPG utility (main for gpg)
+/* g10.c - The GnuPG utility (main for gpg)
  *	Copyright (C) 1998 Free Software Foundation, Inc.
  *
  * This file is part of GNUPG.
@@ -41,6 +41,7 @@
 #include "ttyio.h"
 #include "i18n.h"
 #include "status.h"
+#include "g10defs.h"
 
 #ifndef IS_G10MAINT
   #define IS_G10 1
@@ -87,6 +88,7 @@ enum cmd_and_opt_values { aNull = 0,
     aPrintMD,
     aPrintMDs,
     aCheckTrustDB,
+    aUpdateTrustDB,
     aFixTrustDB,
     aListTrustDB,
     aListTrustPath,
@@ -97,7 +99,6 @@ enum cmd_and_opt_values { aNull = 0,
     aGenRandom,
 
     oFingerprint,
-    oDoNotExportRSA,
     oAnswerYes,
     oAnswerNo,
     oKeyring,
@@ -136,6 +137,7 @@ enum cmd_and_opt_values { aNull = 0,
     oSetFilename,
     oComment,
     oThrowKeyid,
+    oForceV3Sigs,
     oS2KMode,
     oS2KDigest,
     oS2KCipher,
@@ -171,13 +173,17 @@ static ARGPARSE_OPTS opts[] = {
   #endif
     { aExport, "export"          , 256, N_("export keys") },
     { aExportSecret, "export-secret-keys" , 256, "@" },
-    { oDoNotExportRSA, "do-not-export-rsa", 0, "@" },
     { aImport, "import",      256     , N_("import/merge keys")},
     { aListPackets, "list-packets",256,N_("list only the sequence of packets")},
   #ifdef IS_G10MAINT
-    { aExportOwnerTrust, "export-ownertrust", 256, N_("export the ownertrust values")},
-    { aImportOwnerTrust, "import-ownertrust", 256 , N_("import ownertrust values")},
-    { aCheckTrustDB, "check-trustdb",0 , N_("|[NAMES]|check the trust database")},
+    { aExportOwnerTrust,
+	      "export-ownertrust", 256, N_("export the ownertrust values")},
+    { aImportOwnerTrust,
+	      "import-ownertrust", 256 , N_("import ownertrust values")},
+    { aUpdateTrustDB,
+	      "update-trustdb",0 , N_("|[NAMES]|update the trust database")},
+    { aCheckTrustDB,
+	      "check-trustdb",0 , N_("|[NAMES]|check the trust database")},
     { aFixTrustDB, "fix-trustdb",0 , N_("fix a corrupted trust database")},
     { aDeArmor, "dearmor", 256, N_("De-Armor a file or stdin") },
     { aEnArmor, "enarmor", 256, N_("En-Armor a file or stdin") },
@@ -200,6 +206,7 @@ static ARGPARSE_OPTS opts[] = {
   #endif
     { oOutput, "output",    2, N_("use as output file")},
     { oVerbose, "verbose",   0, N_("verbose") },
+    { oForceV3Sigs, "force-v3-sigs", 0, N_("force v3 signatures") },
  /* { oDryRun, "dry-run",   0, N_("do not make any changes") }, */
     { oBatch, "batch",     0, N_("batch mode: never ask")},
     { oAnswerYes, "yes",       0, N_("assume yes on most questions")},
@@ -301,9 +308,9 @@ strusage( int level )
     switch( level ) {
       case 11: p =
 	  #ifdef IS_G10MAINT
-	    "gpgm (GNUPG)";
+	    "gpgm (GnuPG)";
 	  #else
-	    "gpg (GNUPG)";
+	    "gpg (GnuPG)";
 	  #endif
 	break;
       case 13: p = VERSION; break;
@@ -322,7 +329,7 @@ strusage( int level )
       case 41:	p =
 	  #ifdef IS_G10MAINT
 	    _("Syntax: gpgm [options] [files]\n"
-	      "GNUPG maintenance utility\n");
+	      "GnuPG maintenance utility\n");
 	  #else
 	    _("Syntax: gpg [options] [files]\n"
 	      "sign, check, encrypt or decrypt\n"
@@ -503,8 +510,8 @@ main( int argc, char **argv )
     log_set_name("gpg");
     secure_random_alloc(); /* put random number into secure memory */
     disable_core_dumps();
-    init_signals();
   #endif
+    init_signals();
     i18n_init();
     opt.compress = -1; /* defaults to standard compress level */
     /* fixme: set the next two to zero and decide where used */
@@ -571,7 +578,6 @@ main( int argc, char **argv )
     /* Okay, we are now working under our real uid */
   #endif
 
-
     if( default_config )
 	configname = make_filename(opt.homedir, "options", NULL );
 
@@ -638,6 +644,7 @@ main( int argc, char **argv )
 	  case aPrintMDs: set_cmd( &cmd, aPrintMDs); break;
 	  case aListTrustDB: set_cmd( &cmd, aListTrustDB); break;
 	  case aCheckTrustDB: set_cmd( &cmd, aCheckTrustDB); break;
+	  case aUpdateTrustDB: set_cmd( &cmd, aUpdateTrustDB); break;
 	  case aFixTrustDB: set_cmd( &cmd, aFixTrustDB); break;
 	  case aListTrustPath: set_cmd( &cmd, aListTrustPath); break;
 	  case aDeArmor: set_cmd( &cmd, aDeArmor); break;
@@ -692,10 +699,12 @@ main( int argc, char **argv )
 	  case oCompressKeys: opt.compress_keys = 1; break;
 	  case aListSecretKeys: set_cmd( &cmd, aListSecretKeys); break;
 	  case oAlwaysTrust: opt.always_trust = 1; break;
-	  case oLoadExtension: register_cipher_extension(pargs.r.ret_str); break;
+	  case oLoadExtension:
+	    register_cipher_extension(orig_argc? *orig_argv:NULL,
+				      pargs.r.ret_str);
+	    break;
 	  case oRFC1991: opt.rfc1991 = 1; opt.no_comment = 1; break;
 	  case oEmuChecksumBug: opt.emulate_bugs |= EMUBUG_GPGCHKSUM; break;
-	  case oDoNotExportRSA: opt.do_not_export_rsa = 1; break;
 	  case oCompressSigs: opt.compress_sigs = 1; break;
 	  case oRunAsShmCP:
 	  #ifndef USE_SHM_COPROCESSING
@@ -707,6 +716,7 @@ main( int argc, char **argv )
 	  case oSetFilename: opt.set_filename = pargs.r.ret_str; break;
 	  case oComment: opt.comment_string = pargs.r.ret_str; break;
 	  case oThrowKeyid: opt.throw_keyid = 1; break;
+	  case oForceV3Sigs: opt.force_v3_sigs = 1; break;
 	  case oS2KMode:   opt.s2k_mode = pargs.r.ret_int; break;
 	  case oS2KDigest: s2k_digest_string = m_strdup(pargs.r.ret_str); break;
 	  case oS2KCipher: s2k_cipher_string = m_strdup(pargs.r.ret_str); break;
@@ -835,13 +845,13 @@ main( int argc, char **argv )
 	&& !(cmd == aKMode && argc == 2 ) ) {
 
 	if( !sec_nrings || default_keyring )  /* add default secret rings */
-	    add_secret_keyring("secring.gpg");
+	    add_keyblock_resource("secring.gpg", 0, 1);
 	for(sl = sec_nrings; sl; sl = sl->next )
-	    add_secret_keyring( sl->d );
+	    add_keyblock_resource( sl->d, 0, 1 );
 	if( !nrings || default_keyring )  /* add default ring */
-	    add_keyring("pubring.gpg");
+	    add_keyblock_resource("pubring.gpg", 0, 0);
 	for(sl = nrings; sl; sl = sl->next )
-	    add_keyring( sl->d );
+	    add_keyblock_resource( sl->d, 0, 0 );
     }
     FREE_STRLIST(nrings);
     FREE_STRLIST(sec_nrings);
@@ -996,7 +1006,7 @@ main( int argc, char **argv )
 	    else {
 		/* add keyring (default keyrings are not registered in this
 		 * special case */
-		add_keyring( argv[1] );
+		add_keyblock_resource( argv[1], 0, 0 );
 		public_key_list( **argv?1:0, argv );
 	    }
 	}
@@ -1160,6 +1170,12 @@ main( int argc, char **argv )
 	}
 	break;
 
+      case aUpdateTrustDB:
+	if( argc )
+	    wrong_args("--update-trustdb");
+	update_trustdb();
+	break;
+
       case aCheckTrustDB:
 	if( !argc )
 	    check_trustdb(NULL);
@@ -1232,6 +1248,8 @@ main( int argc, char **argv )
 void
 g10_exit( int rc )
 {
+    if( opt.debug & DBG_MEMSTAT_VALUE )
+	m_print_stats("on exit");
     if( opt.debug )
 	secmem_dump_stats();
     secmem_term();
