@@ -33,41 +33,53 @@
 #include "main.h"
 #include "i18n.h"
 
-#define MAX_PK_CACHE_ENTRIES 500
+#define MAX_UNK_CACHE_ENTRIES 1000
+#define MAX_PK_CACHE_ENTRIES	50
+#define MAX_UID_CACHE_ENTRIES	50
+
+static struct {
+    int any;
+    int okay_count;
+    int nokey_count;
+    int error_count;
+} lkup_stats[21];
 
 
-typedef struct keyid_list {
-    struct keyid_list *next;
-    u32 keyid[2];
-} *keyid_list_t;
 
+
+#if MAX_UNK_CACHE_ENTRIES
+  typedef struct keyid_list {
+      struct keyid_list *next;
+      u32 keyid[2];
+  } *keyid_list_t;
+  static keyid_list_t unknown_keyids;
+  static int unk_cache_entries;   /* number of entries in unknown keys cache */
+  static int unk_cache_disabled;
+#endif
+
+#if MAX_PK_CACHE_ENTRIES
+  typedef struct pk_cache_entry {
+      struct pk_cache_entry *next;
+      u32 keyid[2];
+      PKT_public_key *pk;
+  } *pk_cache_entry_t;
+  static pk_cache_entry_t pk_cache;
+  static int pk_cache_entries;	 /* number of entries in pk cache */
+  static int pk_cache_disabled;
+#endif
+
+#if MAX_UID_CACHE_ENTRIES < 5
+    #error we really need the userid cache
+#endif
 typedef struct user_id_db {
     struct user_id_db *next;
     u32 keyid[2];
     int len;
     char name[1];
 } *user_id_db_t;
-
-typedef struct pk_cache_entry {
-    struct pk_cache_entry *next;
-    u32 keyid[2];
-    PKT_public_key *pk;
-} *pk_cache_entry_t;
-
-typedef struct enum_seckey_context {
-    int eof;
-    STRLIST sl;
-    IOBUF iobuf;
-} enum_seckey_context_t;
-
-
-static STRLIST keyrings;
-static STRLIST secret_keyrings;
-
-static keyid_list_t unknown_keyids;
 static user_id_db_t user_id_db;
-static pk_cache_entry_t pk_cache;
-static int pk_cache_entries;   /* number of entries in pk cache */
+static int uid_cache_entries;	/* number of entries in uid cache */
+
 
 
 static int lookup( PKT_public_key *pk,
@@ -76,99 +88,35 @@ static int lookup( PKT_public_key *pk,
 static int lookup_sk( PKT_secret_key *sk,
 		   int mode,  u32 *keyid, const char *name, int primary );
 
-/* note this function may be called before secure memory is
- * available
- * The first keyring which is added by this function is
- * created if it does not exist.
- */
-void
-add_keyring( const char *name )
+
+
+static void
+print_stats()
 {
-    STRLIST sl;
-    int rc, force = !keyrings;
-
-    if( *name != '/' ) { /* do tilde expansion etc */
-	char *p ;
-
-	if( strchr(name, '/') )
-	    p = make_filename(name, NULL);
-	else
-	    p = make_filename(opt.homedir, name, NULL);
-	sl = append_to_strlist( &keyrings, p );
-	m_free(p);
+    int i;
+    for(i=0; i < DIM(lkup_stats); i++ ) {
+	if( lkup_stats[i].any )
+	    fprintf(stderr,
+		    "lookup stats: mode=%-2d  ok=%-6d  nokey=%-6d  err=%-6d\n",
+		    i,
+		    lkup_stats[i].okay_count,
+		    lkup_stats[i].nokey_count,
+		    lkup_stats[i].error_count );
     }
-    else
-	sl = append_to_strlist( &keyrings, name );
-
-    /* fixme: We should remove much out of this module and
-     * combine it with the keyblock stuff from ringedit.c
-     * For now we will simple add the filename as keyblock resource
-     */
-    rc = add_keyblock_resource( sl->d, force, 0 );
-    if( rc )
-	log_error("keyblock resource '%s': %s\n", sl->d, g10_errstr(rc) );
 }
 
 
-/****************
- * Get the name of the keyrings, start with a sequence number of 0.
- */
-const char *
-get_keyring( int sequence )
-{
-    STRLIST sl;
-
-    for(sl = keyrings; sl && sequence; sl = sl->next, sequence-- )
-	;
-    return sl? sl->d : NULL;
-}
-
-const char *
-get_secret_keyring( int sequence )
-{
-    STRLIST sl;
-
-    for(sl = secret_keyrings; sl && sequence; sl = sl->next, sequence-- )
-	;
-    return sl? sl->d : NULL;
-}
-
-
-void
-add_secret_keyring( const char *name )
-{
-    STRLIST sl;
-    int rc, force = !secret_keyrings;
-
-    if( *name != '/' ) { /* do tilde expansion etc */
-	char *p ;
-
-	if( strchr(name, '/') )
-	    p = make_filename(name, NULL);
-	else
-	    p = make_filename(opt.homedir, name, NULL);
-	sl = append_to_strlist( &secret_keyrings, p );
-	m_free(p);
-    }
-    else
-	sl = append_to_strlist( &secret_keyrings, name );
-
-    /* fixme: We should remove much out of this module and
-     * combine it with the keyblock stuff from ringedit.c
-     * For now we will simple add the filename as keyblock resource
-     */
-    rc = add_keyblock_resource( sl->d, force, 1 );
-    if( rc )
-	log_error("secret keyblock resource '%s': %s\n", sl->d, g10_errstr(rc));
-}
-
 
 static void
 cache_public_key( PKT_public_key *pk )
 {
+  #if MAX_PK_CACHE_ENTRIES
     pk_cache_entry_t ce;
     u32 keyid[2];
 
+    if( pk_cache_disabled )
+	return;
+
     if( is_ELGAMAL(pk->pubkey_algo)
 	|| pk->pubkey_algo == PUBKEY_ALGO_DSA
 	|| is_RSA(pk->pubkey_algo) ) {
@@ -184,24 +132,20 @@ cache_public_key( PKT_public_key *pk )
 	    return;
 	}
 
-    if( pk_cache_entries > MAX_PK_CACHE_ENTRIES ) {
-	/* FIMXE: use another algorithm to free some cache slots */
-	if( pk_cache_entries == MAX_PK_CACHE_ENTRIES )	{
-	    pk_cache_entries++;
-	    log_info("too many entries in pk cache - disabled\n");
-	}
-	ce = pk_cache;
-	free_public_key( ce->pk );
-    }
-    else {
-	pk_cache_entries++;
-	ce = m_alloc( sizeof *ce );
-	ce->next = pk_cache;
-	pk_cache = ce;
+    if( pk_cache_entries >= MAX_PK_CACHE_ENTRIES ) {
+	/* fixme: use another algorithm to free some cache slots */
+	pk_cache_disabled=1;
+	log_info("too many entries in pk cache - disabled\n");
+	return;
     }
+    pk_cache_entries++;
+    ce = m_alloc( sizeof *ce );
+    ce->next = pk_cache;
+    pk_cache = ce;
     ce->pk = copy_public_key( NULL, pk );
     ce->keyid[0] = keyid[0];
     ce->keyid[1] = keyid[1];
+  #endif
 }
 
 
@@ -220,6 +164,13 @@ cache_user_id( PKT_user_id *uid, u32 *keyid )
 	    return;
 	}
 
+    if( uid_cache_entries >= MAX_UID_CACHE_ENTRIES ) {
+	/* fixme: use another algorithm to free some cache slots */
+	r = user_id_db;
+	user_id_db = r->next;
+	m_free(r);
+	uid_cache_entries--;
+    }
     r = m_alloc( sizeof *r + uid->len-1 );
     r->keyid[0] = keyid[0];
     r->keyid[1] = keyid[1];
@@ -227,6 +178,7 @@ cache_user_id( PKT_user_id *uid, u32 *keyid )
     memcpy(r->name, uid->name, r->len);
     r->next = user_id_db;
     user_id_db = r;
+    uid_cache_entries++;
 }
 
 
@@ -239,23 +191,30 @@ cache_user_id( PKT_user_id *uid, u32 *keyid )
 int
 get_pubkey( PKT_public_key *pk, u32 *keyid )
 {
-    keyid_list_t kl;
     int internal = 0;
     int rc = 0;
-    pk_cache_entry_t ce;
 
-    /* let's see whether we checked the keyid already */
-    for( kl = unknown_keyids; kl; kl = kl->next )
-	if( kl->keyid[0] == keyid[0] && kl->keyid[1] == keyid[1] )
-	    return G10ERR_NO_PUBKEY; /* already checked and not found */
+  #if MAX_UNK_CACHE_ENTRIES
+    {	/* let's see whether we checked the keyid already */
+	keyid_list_t kl;
+	for( kl = unknown_keyids; kl; kl = kl->next )
+	    if( kl->keyid[0] == keyid[0] && kl->keyid[1] == keyid[1] )
+		return G10ERR_NO_PUBKEY; /* already checked and not found */
+    }
+  #endif
 
-    /* Try to get it from our cache */
-    for( ce = pk_cache; ce; ce = ce->next )
-	if( ce->keyid[0] == keyid[0] && ce->keyid[1] == keyid[1] ) {
-	    if( pk )
-		copy_public_key( pk, ce->pk );
-	    return 0;
+  #if MAX_PK_CACHE_ENTRIES
+    {	/* Try to get it from the cache */
+	pk_cache_entry_t ce;
+	for( ce = pk_cache; ce; ce = ce->next ) {
+	    if( ce->keyid[0] == keyid[0] && ce->keyid[1] == keyid[1] ) {
+		if( pk )
+		    copy_public_key( pk, ce->pk );
+		return 0;
+	    }
 	}
+    }
+  #endif
     /* more init stuff */
     if( !pk ) {
 	pk = m_alloc_clear( sizeof *pk );
@@ -268,12 +227,24 @@ get_pubkey( PKT_public_key *pk, u32 *keyid )
     if( !rc )
 	goto leave;
 
+  #if MAX_UNK_CACHE_ENTRIES
     /* not found: store it for future reference */
-    kl = m_alloc( sizeof *kl );
-    kl->keyid[0] = keyid[0];
-    kl->keyid[1] = keyid[1];
-    kl->next = unknown_keyids;
-    unknown_keyids = kl;
+    if( unk_cache_disabled )
+	;
+    else if( ++unk_cache_entries > MAX_UNK_CACHE_ENTRIES ) {
+	unk_cache_disabled = 1;
+	log_info("too many entries in unk cache - disabled\n");
+    }
+    else {
+	keyid_list_t kl;
+
+	kl = m_alloc( sizeof *kl );
+	kl->keyid[0] = keyid[0];
+	kl->keyid[1] = keyid[1];
+	kl->next = unknown_keyids;
+	unknown_keyids = kl;
+    }
+  #endif
     rc = G10ERR_NO_PUBKEY;
 
   leave:
@@ -639,7 +610,7 @@ merge_one_pk_and_selfsig( KBNODE keyblock, KBNODE knode )
     for(k=keyblock; k; k = k->next ) {
 	if( k->pkt->pkttype == PKT_SIGNATURE
 	    && (sig=k->pkt->pkt.signature)->sig_class >= 0x10
-	    && sig->sig_class <= 0x13
+	    && sig->sig_class <= 0x30
 	    && sig->keyid[0] == kid[0]
 	    && sig->keyid[1] == kid[1]
 	    && sig->version > 3 ) {
@@ -650,7 +621,7 @@ merge_one_pk_and_selfsig( KBNODE keyblock, KBNODE knode )
 	     */
 	    const byte *p;
 	    p = parse_sig_subpkt( sig->hashed_data, SIGSUBPKT_KEY_EXPIRE, NULL );
-	    pk->valid_days = p? ((buffer_to_u32(p)+86399L)/86400L):0;
+	    pk->expiredate = p? buffer_to_u32(p):0;
 	    /* fixme: add usage etc. to pk */
 	    break;
 	}
@@ -689,7 +660,7 @@ merge_keys_and_selfsig( KBNODE keyblock )
 	}
 	else if( (pk || sk ) && k->pkt->pkttype == PKT_SIGNATURE
 		 && (sig=k->pkt->pkt.signature)->sig_class >= 0x10
-		 && sig->sig_class <= 0x13 && sig->version > 3
+		 && sig->sig_class <= 0x30 && sig->version > 3
 		 && sig->keyid[0] == kid[0] && sig->keyid[1] == kid[1] ) {
 	    /* okay this is (the first) self-signature which can be used
 	     * FIXME: We should only use this if the signature is valid
@@ -699,12 +670,12 @@ merge_keys_and_selfsig( KBNODE keyblock )
 	    const byte *p;
 	    p = parse_sig_subpkt( sig->hashed_data, SIGSUBPKT_KEY_EXPIRE, NULL );
 	    if( pk ) {
-		pk->valid_days = p? ((buffer_to_u32(p)+86399L)/86400L):0;
+		pk->expiredate = p? buffer_to_u32(p):0;
 		/* fixme: add usage etc. */
 		pk = NULL; /* use only the first self signature */
 	    }
 	    else {
-		sk->valid_days = p? ((buffer_to_u32(p)+86399L)/86400L):0;
+		sk->expiredate = p? buffer_to_u32(p):0;
 		sk = NULL; /* use only the first self signature */
 	    }
 	}
@@ -716,7 +687,7 @@ merge_keys_and_selfsig( KBNODE keyblock )
 
 
 /****************
- * Lookup a key by scanning all keyrings
+ * Lookup a key by scanning all keyresources
  *   mode 1 = lookup by NAME (exact)
  *	  2 = lookup by NAME (substring)
  *	  3 = lookup by NAME (email address)
@@ -829,9 +800,10 @@ lookup( PKT_public_key *pk, int mode,  u32 *keyid,
 			    break;
 		    }
 		    else if( mode == 16 || mode == 20 ) {
+			byte afp[MAX_FINGERPRINT_LEN];
 			size_t an;
-			byte *afp = fingerprint_from_pk(
-					k->pkt->pkt.public_key, NULL, &an );
+
+			fingerprint_from_pk(k->pkt->pkt.public_key, afp, &an );
 
 			if( DBG_CACHE ) {
 			    u32 aki[2];
@@ -845,10 +817,8 @@ lookup( PKT_public_key *pk, int mode,  u32 *keyid,
 			    && ( !pk->pubkey_algo
 				 || pk->pubkey_algo
 				    == k->pkt->pkt.public_key->pubkey_algo) ) {
-			    m_free(afp);
 			    break;
 			}
-			m_free(afp);
 		    }
 		    else
 			BUG();
@@ -910,6 +880,24 @@ lookup( PKT_public_key *pk, int mode,  u32 *keyid,
     enum_keyblocks( 2, &kbpos, &keyblock ); /* close */
     release_kbnode( keyblock );
     set_packet_list_mode(oldmode);
+    if( opt.debug & DBG_MEMSTAT_VALUE ) {
+	static int initialized;
+
+	if( !initialized ) {
+	    initialized = 1;
+	    atexit( print_stats );
+	}
+
+	assert( mode < DIM(lkup_stats) );
+	lkup_stats[mode].any = 1;
+	if( !rc )
+	    lkup_stats[mode].okay_count++;
+	else if ( rc == G10ERR_NO_PUBKEY )
+	    lkup_stats[mode].nokey_count++;
+	else
+	    lkup_stats[mode].error_count++;
+    }
+
     return rc;
 }
 
@@ -1008,16 +996,15 @@ lookup_sk( PKT_secret_key *sk, int mode,  u32 *keyid, const char *name,
 		    }
 		    else if( mode == 16 || mode == 20 ) {
 			size_t an;
-			byte *afp = fingerprint_from_sk(
-					k->pkt->pkt.secret_key, NULL, &an );
+			byte afp[MAX_FINGERPRINT_LEN];
+
+			fingerprint_from_sk(k->pkt->pkt.secret_key, afp, &an );
 			if( an == mode && !memcmp( afp, name, an)
 			    && ( !sk->pubkey_algo
 				 || sk->pubkey_algo
 				    == k->pkt->pkt.secret_key->pubkey_algo) ) {
-			    m_free(afp);
 			    break;
 			}
-			m_free(afp);
 		    }
 		    else
 			BUG();
@@ -1069,12 +1056,19 @@ enum_secret_keys( void **context, PKT_secret_key *sk, int with_subkeys )
     int rc=0;
     PACKET pkt;
     int save_mode;
-    enum_seckey_context_t *c = *context;
+    struct {
+	int eof;
+	int sequence;
+	const char *name;
+	IOBUF iobuf;
+    } *c = *context;
+
 
     if( !c ) { /* make a new context */
 	c = m_alloc_clear( sizeof *c );
 	*context = c;
-	c->sl = secret_keyrings;
+	c->sequence = 0;
+	c->name = enum_keyblock_resources( &c->sequence, 1 );
     }
 
     if( !sk ) { /* free the context */
@@ -1088,10 +1082,11 @@ enum_secret_keys( void **context, PKT_secret_key *sk, int with_subkeys )
     if( c->eof )
 	return -1;
 
-    for( ; c->sl; c->sl = c->sl->next ) {
+    /* FIXME: This assumes a plain keyring file */
+    for( ; c->name; c->name = enum_keyblock_resources( &c->sequence, 1 ) ) {
 	if( !c->iobuf ) {
-	    if( !(c->iobuf = iobuf_open( c->sl->d ) ) ) {
-		log_error("enum_secret_keys: can't open '%s'\n", c->sl->d );
+	    if( !(c->iobuf = iobuf_open( c->name ) ) ) {
+		log_error("enum_secret_keys: can't open '%s'\n", c->name );
 		continue; /* try next file */
 	    }
 	}
@@ -1127,7 +1122,7 @@ get_user_id_string( u32 *keyid )
     user_id_db_t r;
     char *p;
     int pass=0;
-    /* try it two times; second pass reads from keyrings */
+    /* try it two times; second pass reads from key resources */
     do {
 	for(r=user_id_db; r; r = r->next )
 	    if( r->keyid[0] == keyid[0] && r->keyid[1] == keyid[1] ) {
@@ -1147,7 +1142,7 @@ get_user_id( u32 *keyid, size_t *rn )
     user_id_db_t r;
     char *p;
     int pass=0;
-    /* try it two times; second pass reads from keyrings */
+    /* try it two times; second pass reads from key resources */
     do {
 	for(r=user_id_db; r; r = r->next )
 	    if( r->keyid[0] == keyid[0] && r->keyid[1] == keyid[1] ) {
