@@ -71,12 +71,16 @@ int clearsign_file( const char *fname, STRLIST locusr, const char *outfile );
 /*-- sig-check.c --*/
 int check_key_signature( KBNODE root, KBNODE node, int *is_selfsig );
 
-/*-- keyedit.c --*/
+/*-- delkey.c --*/
 int delete_key( const char *username, int secure );
+
+/*-- keyedit.c --*/
 void keyedit_menu( const char *username, STRLIST locusr );
 
 /*-- keygen.c --*/
+u32 ask_expiredate(void);
 void generate_keypair(void);
+int keygen_add_key_expire( PKT_signature *sig, void *opaque );
 int keygen_add_std_prefs( PKT_signature *sig, void *opaque );
 int generate_subkeypair( KBNODE pub_keyblock, KBNODE sec_keyblock );
 
