@@ -73,7 +73,6 @@ do_check( PKT_secret_key *sk )
 	    int ndata;
 	    byte *p, *data;
 
-
 	    i = pubkey_get_npkey(sk->pubkey_algo);
 	    assert( mpi_is_opaque( sk->skey[i] ) );
 	    p = mpi_get_opaque( sk->skey[i], &ndata );
@@ -154,19 +153,22 @@ do_check( PKT_secret_key *sk )
 
 /****************
  * Check the secret key
- * Ask up to 3 times for a correct passphrase
+ * Ask up to 3 (or n) times for a correct passphrase
  */
 int
-check_secret_key( PKT_secret_key *sk )
+check_secret_key( PKT_secret_key *sk, int n )
 {
     int rc = G10ERR_BAD_PASS;
     int i;
 
-    for(i=0; i < 3 && rc == G10ERR_BAD_PASS; i++ ) {
+    if( n < 1 )
+	n = 3; /* use the default value */
+
+    for(i=0; i < n && rc == G10ERR_BAD_PASS; i++ ) {
 	if( i )
 	    log_error(_("Invalid passphrase; please try again ...\n"));
 	rc = do_check( sk );
-	if( get_passphrase_fd() != -1 )
+	if( have_static_passphrase() )
 	    break;
     }
 
@@ -209,7 +211,9 @@ protect_secret_key( PKT_secret_key *sk, DEK *dek )
 	else {
 	    cipher_hd = cipher_open( sk->protect.algo,
 				     CIPHER_MODE_AUTO_CFB, 1 );
-	    cipher_setkey( cipher_hd, dek->key, dek->keylen );
+	    if( cipher_setkey( cipher_hd, dek->key, dek->keylen ) )
+		log_info(_("Warning: Weak key detected"
+			   " - please change passphrase again.\n"));
 	    cipher_setiv( cipher_hd, NULL );
 	    cipher_encrypt( cipher_hd, sk->protect.iv, sk->protect.iv, 8 );
 	    if( sk->version >= 4 ) {
