@@ -40,6 +40,39 @@
 #include "tdbio.h"
 
 
+/****************
+ * Yes, this is a very simple implementation. We should really
+ * use a page aligned buffer and read complete pages.
+ * To implement a simple trannsaction system, this is sufficient.
+ */
+typedef struct cache_ctrl_struct *CACHE_CTRL;
+struct cache_ctrl_struct {
+    CACHE_CTRL next;
+    struct {
+	unsigned used:1;
+	unsigned dirty:1;
+    } flags;
+    ulong recno;
+    char data[TRUST_RECORD_LEN];
+};
+
+#define MAX_CACHE_ENTRIES    200
+static CACHE_CTRL cache_list;
+static int cache_entries;
+
+/* a type used to pass infomation to cmp_krec_fpr */
+struct cmp_krec_fpr_struct {
+    int pubkey_algo;
+    const char *fpr;
+    int fprlen;
+};
+
+/* a type used to pass infomation to cmp_sdir */
+struct cmp_sdir_struct {
+    int pubkey_algo;
+    u32 keyid[2];
+};
+
 
 static char *db_name;
 static int  db_fd = -1;
@@ -47,6 +80,184 @@ static int  db_fd = -1;
 
 static void open_db(void);
 
+
+/*************************************
+ ************* record cache **********
+ *************************************/
+
+/****************
+ * Get the data from therecord cache and return a
+ * pointer into that cache.  Caller should copy
+ * the return data.  NULL is returned on a cache miss.
+ */
+static const char *
+get_record_from_cache( ulong recno )
+{
+    CACHE_CTRL r;
+
+    for( r = cache_list; r; r = r->next ) {
+	if( r->flags.used && r->recno == recno )
+	    return r->data;
+    }
+    return NULL;
+}
+
+
+static int
+write_cache_item( CACHE_CTRL r )
+{
+    int n;
+
+    if( lseek( db_fd, r->recno * TRUST_RECORD_LEN, SEEK_SET ) == -1 ) {
+	log_error(_("trustdb rec %lu: lseek failed: %s\n"),
+					    r->recno, strerror(errno) );
+	return G10ERR_WRITE_FILE;
+    }
+    n = write( db_fd, r->data, TRUST_RECORD_LEN);
+    if( n != TRUST_RECORD_LEN ) {
+	log_error(_("trustdb rec %lu: write failed (n=%d): %s\n"),
+					    r->recno, n, strerror(errno) );
+	return G10ERR_WRITE_FILE;
+    }
+    r->flags.dirty = 0;
+    return 0;
+}
+
+/****************
+ * Put data into the cache.  This function may flush the
+ * some cache entries if there is not enough space available.
+ */
+int
+put_record_into_cache( ulong recno, const char *data )
+{
+    CACHE_CTRL r, unused;
+    int dirty_count = 0;
+    int clean_count = 0;
+
+    /* see whether we already cached this one */
+    for( unused = NULL, r = cache_list; r; r = r->next ) {
+	if( !r->flags.used ) {
+	    if( !unused )
+		unused = r;
+	}
+	else if( r->recno == recno ) {
+	    if( !r->flags.dirty ) {
+		/* Hmmm: should we use a a copy and compare? */
+		if( memcmp(r->data, data, TRUST_RECORD_LEN ) )
+		    r->flags.dirty = 1;
+	    }
+	    memcpy( r->data, data, TRUST_RECORD_LEN );
+	    return 0;
+	}
+	if( r->flags.used ) {
+	    if( r->flags.dirty )
+		dirty_count++;
+	    else
+		clean_count++;
+	}
+    }
+    /* not in the cache: add a new entry */
+    if( unused ) { /* reuse this entry */
+	r = unused;
+	r->flags.used = 1;
+	r->recno = recno;
+	memcpy( r->data, data, TRUST_RECORD_LEN );
+	r->flags.dirty = 1;
+	cache_entries++;
+	return 0;
+    }
+    /* see whether we reached the limit */
+    if( cache_entries < MAX_CACHE_ENTRIES ) { /* no */
+	r = m_alloc( sizeof *r );
+	r->flags.used = 1;
+	r->recno = recno;
+	memcpy( r->data, data, TRUST_RECORD_LEN );
+	r->flags.dirty = 1;
+	r->next = cache_list;
+	cache_list = r;
+	cache_entries++;
+	return 0;
+    }
+    /* cache is full: discard some clean entries */
+    if( clean_count ) {
+	int n = clean_count / 3; /* discard a third of the clean entries */
+	if( !n )
+	    n = 1;
+	for( unused = NULL, r = cache_list; r; r = r->next ) {
+	    if( r->flags.used && !r->flags.dirty ) {
+		if( !unused )
+		    unused = r;
+		r->flags.used = 0;
+		cache_entries--;
+		if( !--n )
+		    break;
+	    }
+	}
+	assert( unused );
+	r = unused;
+	r->flags.used = 1;
+	r->recno = recno;
+	memcpy( r->data, data, TRUST_RECORD_LEN );
+	r->flags.dirty = 1;
+	cache_entries++;
+	return 0;
+    }
+    /* no clean entries: have to flush some dirty entries */
+    if( dirty_count ) {
+	int n = dirty_count / 5; /* discard some dirty entries */
+	if( !n )
+	    n = 1;
+	for( unused = NULL, r = cache_list; r; r = r->next ) {
+	    if( r->flags.used && r->flags.dirty ) {
+		int rc = write_cache_item( r );
+		if( rc )
+		    return rc;
+		if( !unused )
+		    unused = r;
+		r->flags.used = 0;
+		cache_entries--;
+		if( !--n )
+		    break;
+	    }
+	}
+	assert( unused );
+	r = unused;
+	r->flags.used = 1;
+	r->recno = recno;
+	memcpy( r->data, data, TRUST_RECORD_LEN );
+	r->flags.dirty = 1;
+	cache_entries++;
+	return 0;
+    }
+    BUG();
+}
+
+
+
+/****************
+ * Sync the cache to disk
+ */
+
+int
+tdbio_sync()
+{
+    CACHE_CTRL r;
+
+    for( r = cache_list; r; r = r->next ) {
+	if( r->flags.used && r->flags.dirty ) {
+	    int rc = write_cache_item( r );
+	    if( rc )
+		return rc;
+	}
+    }
+    return 0;
+}
+
+
+
+/********************************************************
+ **************** cached I/O functions ******************
+ ********************************************************/
 
 int
 tdbio_set_dbname( const char *new_dbname, int create )
@@ -106,6 +317,8 @@ tdbio_set_dbname( const char *new_dbname, int create )
 	    rec.rectype = RECTYPE_VER;
 	    rec.recnum = 0;
 	    rc = tdbio_write_record( &rec );
+	    if( !rc )
+		tdbio_sync();
 	    if( rc )
 		log_fatal_f( fname, _("failed to create version record: %s"),
 							       g10_errstr(rc));
@@ -148,6 +361,49 @@ open_db()
 }
 
 
+/****************
+ * Make a hashtable: type 0 = key hash, 1 = sdir hash
+ */
+static void
+create_hashtable( TRUSTREC *vr, int type )
+{
+    TRUSTREC rec;
+    off_t offset;
+    ulong recnum;
+    int i, n, rc;
+
+    offset = lseek( db_fd, 0, SEEK_END );
+    if( offset == -1 )
+	log_fatal("trustdb: lseek to end failed: %s\n", strerror(errno) );
+    recnum = offset / TRUST_RECORD_LEN;
+    assert(recnum); /* this is will never be the first record */
+
+    if( !type )
+	vr->r.ver.keyhashtbl = recnum;
+    else
+	vr->r.ver.sdirhashtbl = recnum;
+    /* Now write the records */
+    n = (256+ITEMS_PER_HTBL_RECORD-1) / ITEMS_PER_HTBL_RECORD;
+    for(i=0; i < n; i++, recnum++ ) {
+	 memset( &rec, 0, sizeof rec );
+	 rec.rectype = RECTYPE_HTBL;
+	 rec.recnum = recnum;
+	 rc = tdbio_write_record( &rec );
+	 if( rc )
+	     log_fatal_f(db_name,_("failed to create hashtable: %s\n"),
+						 g10_errstr(rc));
+    }
+    /* update the version record */
+    rc = tdbio_write_record( vr );
+    if( !rc )
+	rc = tdbio_sync();
+    if( rc )
+	log_fatal_f( db_name, _("error updating version record: %s\n"),
+							     g10_errstr(rc));
+}
+
+
+
 /****************
  * Return the record number of the keyhash tbl or create a new one.
  */
@@ -155,58 +411,56 @@ static ulong
 get_keyhashrec()
 {
     static ulong keyhashtbl; /* record number of the key hashtable */
-    TRUSTREC vr;
-    int rc;
 
-    if( keyhashtbl )
-	return keyhashtbl;
+    if( !keyhashtbl ) {
+	TRUSTREC vr;
+	int rc;
 
-    rc = tdbio_read_record( 0, &vr, RECTYPE_VER );
-    if( rc )
-	log_fatal_f( db_name, _("error reading version record: %s\n"),
-							g10_errstr(rc) );
-    if( vr.r.ver.keyhashtbl )
-	keyhashtbl = vr.r.ver.keyhashtbl;
-    else {
-	TRUSTREC rec;
-	off_t offset;
-	ulong recnum;
-	int i, n;
-
-	offset = lseek( db_fd, 0, SEEK_END );
-	if( offset == -1 )
-	    log_fatal("trustdb: lseek to end failed: %s\n", strerror(errno) );
-	recnum = offset / TRUST_RECORD_LEN;
-	assert(recnum); /* this is will never be the first record */
-
-	keyhashtbl = recnum;
-	/* Now write the records */
-	n = (256+ITEMS_PER_HTBL_RECORD-1) / ITEMS_PER_HTBL_RECORD;
-	for(i=0; i < n; i++, recnum++ ) {
-	     memset( &rec, 0, sizeof rec );
-	     rec.rectype = RECTYPE_HTBL; /* free record */
-	     rec.recnum = recnum;
-	     rc = tdbio_write_record( &rec );
-	     if( rc )
-		 log_fatal_f(db_name,_("failed to create hashtable: %s\n"),
-						     g10_errstr(rc));
-	}
-	/* update the version record */
-	vr.r.ver.keyhashtbl = keyhashtbl;
-	rc = tdbio_write_record( &vr );
+	rc = tdbio_read_record( 0, &vr, RECTYPE_VER );
 	if( rc )
-	    log_fatal_f( db_name, _("error updating version record: %s\n"),
-							     g10_errstr(rc));
+	    log_fatal_f( db_name, _("error reading version record: %s\n"),
+							    g10_errstr(rc) );
+	if( !vr.r.ver.keyhashtbl )
+	    create_hashtable( &vr, 0 );
+
+	keyhashtbl = vr.r.ver.keyhashtbl;
     }
     return keyhashtbl;
 }
 
+/****************
+ * Return the record number of the shadow direcory hash table
+ * or create a new one.
+ */
+static ulong
+get_sdirhashrec()
+{
+    static ulong sdirhashtbl; /* record number of the hashtable */
+
+    if( !sdirhashtbl ) {
+	TRUSTREC vr;
+	int rc;
+
+	rc = tdbio_read_record( 0, &vr, RECTYPE_VER );
+	if( rc )
+	    log_fatal_f( db_name, _("error reading version record: %s\n"),
+							    g10_errstr(rc) );
+	if( !vr.r.ver.sdirhashtbl )
+	    create_hashtable( &vr, 1 );
+
+	sdirhashtbl = vr.r.ver.sdirhashtbl;
+    }
+    return sdirhashtbl;
+}
+
 
 /****************
- * Update the key hashtbl or create the table if it does not exist
+ * Update a hashtable.
+ * table gives the start of the table, key and keylen is the key,
+ * newrecnum is the record number to insert.
  */
 static int
-update_keyhashtbl( TRUSTREC *kr )
+upd_hashtable( ulong table, byte *key, int keylen, ulong newrecnum )
 {
     TRUSTREC lastrec, rec;
     ulong hashrec, item;
@@ -214,40 +468,41 @@ update_keyhashtbl( TRUSTREC *kr )
     int level=0;
     int rc, i;
 
-    hashrec = get_keyhashrec();
+    hashrec = table;
   next_level:
-    msb = kr->r.key.fingerprint[level];
+    msb = key[level];
     hashrec += msb / ITEMS_PER_HTBL_RECORD;
     rc = tdbio_read_record( hashrec, &rec, RECTYPE_HTBL );
     if( rc ) {
-	log_error( db_name, "update_keyhashtbl read failed: %s\n",
+	log_error( db_name, "upd_hashtable: read failed: %s\n",
 							g10_errstr(rc) );
 	return rc;
     }
 
     item = rec.r.htbl.item[msb % ITEMS_PER_HTBL_RECORD];
-    if( !item ) { /* insert new one */
-	rec.r.htbl.item[msb % ITEMS_PER_HTBL_RECORD] = kr->recnum;
+    if( !item ) { /* insert a new item into the hash table */
+	rec.r.htbl.item[msb % ITEMS_PER_HTBL_RECORD] = newrecnum;
 	rc = tdbio_write_record( &rec );
 	if( rc ) {
-	    log_error( db_name, "update_keyhashtbl write htbl failed: %s\n",
+	    log_error( db_name, "upd_hashtable: write htbl failed: %s\n",
 							    g10_errstr(rc) );
 	    return rc;
 	}
     }
-    else if( item != kr->recnum ) {  /* must do an update */
+    else if( item != newrecnum ) {  /* must do an update */
 	lastrec = rec;
 	rc = tdbio_read_record( item, &rec, 0 );
 	if( rc ) {
-	    log_error( db_name, "update_keyhashtbl read item failed: %s\n",
+	    log_error( "upd_hashtable: read item failed: %s\n",
 							    g10_errstr(rc) );
 	    return rc;
 	}
+
 	if( rec.rectype == RECTYPE_HTBL ) {
 	    hashrec = item;
 	    level++;
-	    if( level >= kr->r.key.fingerprint_len ) {
-		log_error( db_name, "keyhashtbl has invalid indirections\n");
+	    if( level >= keylen ) {
+		log_error( "hashtable has invalid indirections.\n");
 		return G10ERR_TRUSTDB;
 	    }
 	    goto next_level;
@@ -256,18 +511,15 @@ update_keyhashtbl( TRUSTREC *kr )
 	    /* see whether the key is already in this list */
 	    for(;;) {
 		for(i=0; i < ITEMS_PER_HLST_RECORD; i++ ) {
-		    if( rec.r.hlst.rnum[i] == kr->recnum ) {
-			log_debug("HTBL: no update needed for keyrec %lu\n",
-				    kr->recnum );
-			return 0;
+		    if( rec.r.hlst.rnum[i] == newrecnum ) {
+			return 0; /* okay, already in the list */
 		    }
 		}
 		if( rec.r.hlst.next ) {
 		    rc = tdbio_read_record( rec.r.hlst.next,
-							&rec, RECTYPE_HLST);
+						       &rec, RECTYPE_HLST);
 		    if( rc ) {
-			log_error( db_name,
-				   "scan keyhashtbl read hlst failed: %s\n",
+			log_error( "scan keyhashtbl read hlst failed: %s\n",
 							     g10_errstr(rc) );
 			return rc;
 		    }
@@ -279,21 +531,19 @@ update_keyhashtbl( TRUSTREC *kr )
 	    for(;;) {
 		for(i=0; i < ITEMS_PER_HLST_RECORD; i++ ) {
 		    if( !rec.r.hlst.rnum[i] ) {
-			rec.r.hlst.rnum[i] = kr->recnum;
+			rec.r.hlst.rnum[i] = newrecnum;
 			rc = tdbio_write_record( &rec );
 			if( rc )
-			    log_error( db_name,
-				   "update_keyhashtbl write hlst failed: %s\n",
+			    log_error( "upd_hashtable: write hlst failed: %s\n",
 							      g10_errstr(rc) );
-			return rc; /* ready */
+			return rc; /* done */
 		    }
 		}
 		if( rec.r.hlst.next ) {
 		    rc = tdbio_read_record( rec.r.hlst.next,
 						      &rec, RECTYPE_HLST );
 		    if( rc ) {
-			log_error( db_name,
-				   "update_keyhashtbl read hlst failed: %s\n",
+			log_error( "upd_hashtable: read hlst failed: %s\n",
 							     g10_errstr(rc) );
 			return rc;
 		    }
@@ -302,27 +552,26 @@ update_keyhashtbl( TRUSTREC *kr )
 		    rec.r.hlst.next = item = tdbio_new_recnum();
 		    rc = tdbio_write_record( &rec );
 		    if( rc ) {
-			log_error( db_name,
-			       "update_keyhashtbl write hlst failed: %s\n",
+			log_error( "upd_hashtable: write hlst failed: %s\n",
 							  g10_errstr(rc) );
 			return rc;
 		    }
 		    memset( &rec, 0, sizeof rec );
 		    rec.rectype = RECTYPE_HLST;
 		    rec.recnum = item;
-		    rec.r.hlst.rnum[0] = kr->recnum;
+		    rec.r.hlst.rnum[0] = newrecnum;
+		    rc = tdbio_write_record( &rec );
 		    if( rc )
-			log_error( db_name,
-			       "update_keyhashtbl write ext hlst failed: %s\n",
+			log_error( "upd_hashtable: write ext hlst failed: %s\n",
 							  g10_errstr(rc) );
-		    return rc; /* ready */
+		    return rc; /* done */
 		}
-	    }
+	    } /* end loop over hlst slots */
 	}
-	else if( rec.rectype == RECTYPE_KEY ) { /* insert a list record */
-	    if( rec.recnum == kr->recnum ) {
-		log_debug("HTBL: no update needed for keyrec %lu\n",
-							 kr->recnum );
+	else if( rec.rectype == RECTYPE_KEY
+		 || rec.rectype == RECTYPE_DIR
+		 || rec.rectype == RECTYPE_SDIR ) { /* insert a list record */
+	    if( rec.recnum == newrecnum ) {
 		return 0;
 	    }
 	    item = rec.recnum; /* save number of key record */
@@ -330,11 +579,10 @@ update_keyhashtbl( TRUSTREC *kr )
 	    rec.rectype = RECTYPE_HLST;
 	    rec.recnum = tdbio_new_recnum();
 	    rec.r.hlst.rnum[0] = item;	     /* old keyrecord */
-	    rec.r.hlst.rnum[1] = kr->recnum; /* and new one */
+	    rec.r.hlst.rnum[1] = newrecnum; /* and new one */
 	    rc = tdbio_write_record( &rec );
 	    if( rc ) {
-		log_error( db_name,
-		       "update_keyhashtbl write new hlst failed: %s\n",
+		log_error( "upd_hashtable: write new hlst failed: %s\n",
 						  g10_errstr(rc) );
 		return rc;
 	    }
@@ -342,13 +590,12 @@ update_keyhashtbl( TRUSTREC *kr )
 	    lastrec.r.htbl.item[msb % ITEMS_PER_HTBL_RECORD] = rec.recnum;
 	    rc = tdbio_write_record( &lastrec );
 	    if( rc )
-		log_error( db_name,
-		       "update_keyhashtbl update htbl failed: %s\n",
-						  g10_errstr(rc) );
+		log_error( "upd_hashtable: update htbl failed: %s\n",
+							     g10_errstr(rc) );
 	    return rc; /* ready */
 	}
 	else {
-	    log_error( db_name, "keyhashtbl %lu points to an invalid record\n",
+	    log_error( "hashtbl %lu points to an invalid record\n",
 								    item);
 	    return G10ERR_TRUSTDB;
 	}
@@ -359,6 +606,119 @@ update_keyhashtbl( TRUSTREC *kr )
 
 
 
+/****************
+ * Lookup a record via the hashtable tablewith key/keylen and return the
+ * result in rec.  cmp() should return if the record is the desired one.
+ * Returns -1 if not found, 0 if found or another errocode
+ */
+static int
+lookup_hashtable( ulong table, const byte *key, size_t keylen,
+		  int (*cmpfnc)(void*, const TRUSTREC *), void *cmpdata,
+						TRUSTREC *rec )
+{
+    int rc;
+    ulong hashrec, item;
+    int msb;
+    int level=0;
+
+    hashrec = table;
+  next_level:
+    msb = key[level];
+    hashrec += msb / ITEMS_PER_HTBL_RECORD;
+    rc = tdbio_read_record( hashrec, rec, RECTYPE_HTBL );
+    if( rc ) {
+	log_error( db_name, "lookup_hashtable failed: %s\n", g10_errstr(rc) );
+	return rc;
+    }
+
+    item = rec->r.htbl.item[msb % ITEMS_PER_HTBL_RECORD];
+    if( !item )
+	return -1; /* not found */
+
+    rc = tdbio_read_record( item, rec, 0 );
+    if( rc ) {
+	log_error( db_name, "hashtable read failed: %s\n", g10_errstr(rc) );
+	return rc;
+    }
+    if( rec->rectype == RECTYPE_HTBL ) {
+	hashrec = item;
+	level++;
+	if( level >= keylen ) {
+	    log_error( db_name, "hashtable has invalid indirections\n");
+	    return G10ERR_TRUSTDB;
+	}
+	goto next_level;
+    }
+    else if( rec->rectype == RECTYPE_HLST ) {
+	for(;;) {
+	    int i;
+
+	    for(i=0; i < ITEMS_PER_HLST_RECORD; i++ ) {
+		if( rec->r.hlst.rnum[i] ) {
+		    TRUSTREC tmp;
+
+		    rc = tdbio_read_record( rec->r.hlst.rnum[i], &tmp, 0 );
+		    if( rc ) {
+			log_error( "lookup_hashtable: read item failed: %s\n",
+							      g10_errstr(rc) );
+			return rc;
+		    }
+		    if( (*cmpfnc)( cmpdata, &tmp ) ) {
+			*rec = tmp;
+			return 0;
+		    }
+		}
+	    }
+	    if( rec->r.hlst.next ) {
+		rc = tdbio_read_record( rec->r.hlst.next, rec, RECTYPE_HLST );
+		if( rc ) {
+		    log_error( "lookup_hashtable: read hlst failed: %s\n",
+							 g10_errstr(rc) );
+		    return rc;
+		}
+	    }
+	    else
+		return -1; /* not found */
+	}
+    }
+
+
+    if( (*cmpfnc)( cmpdata, rec ) )
+	return 0; /* really found */
+
+    return -1; /* no: not found */
+}
+
+
+
+
+/****************
+ * Update the key hashtbl or create the table if it does not exist
+ */
+static int
+update_keyhashtbl( TRUSTREC *kr )
+{
+    return upd_hashtable( get_keyhashrec(),
+			  kr->r.key.fingerprint,
+			  kr->r.key.fingerprint_len, kr->recnum );
+}
+
+/****************
+ * Update the shadow dir hashtbl or create the table if it does not exist
+ */
+static int
+update_sdirhashtbl( TRUSTREC *sr )
+{
+    byte key[8];
+
+    u32tobuf( key   , sr->r.sdir.keyid[0] );
+    u32tobuf( key+4 , sr->r.sdir.keyid[1] );
+    return upd_hashtable( get_sdirhashrec(), key, 8, sr->recnum );
+}
+
+
+
+
 void
 tdbio_dump_record( TRUSTREC *rec, FILE *fp  )
 {
@@ -371,8 +731,9 @@ tdbio_dump_record( TRUSTREC *rec, FILE *fp  )
     switch( rec->rectype ) {
       case 0: fprintf(fp, "blank\n");
 	break;
-      case RECTYPE_VER: fprintf(fp, "version, keyhashtbl=%lu, firstfree=%lu\n",
-	    rec->r.ver.keyhashtbl, rec->r.ver.firstfree );
+      case RECTYPE_VER: fprintf(fp, "version, kd=%lu, sd=%lu, free=%lu\n",
+	    rec->r.ver.keyhashtbl, rec->r.ver.sdirhashtbl,
+				   rec->r.ver.firstfree );
 	break;
       case RECTYPE_FREE: fprintf(fp, "free, next=%lu\n", rec->r.free.next );
 	break;
@@ -411,6 +772,10 @@ tdbio_dump_record( TRUSTREC *rec, FILE *fp  )
 		    rec->r.uid.prefrec,
 		    rec->r.uid.siglist,
 		    rec->r.uid.namehash[18], rec->r.uid.namehash[19]);
+	if( rec->r.uid.uidflags & UIDF_CHECKED )
+	    fputs(", checked", fp );
+	if( rec->r.uid.uidflags & UIDF_VALID )
+	    fputs(", valid", fp );
 	if( rec->r.uid.uidflags & UIDF_REVOKED )
 	    fputs(", revoked", fp );
 	putc('\n', fp);
@@ -436,6 +801,14 @@ tdbio_dump_record( TRUSTREC *rec, FILE *fp  )
 	}
 	putc('\n', fp);
 	break;
+      case RECTYPE_SDIR:
+	fprintf(fp, "sdir %lu, keyid=%08lX%08lX, algo=%d, hint=%lu\n",
+		    rec->r.sdir.lid,
+		    (ulong)rec->r.sdir.keyid[0],
+		    (ulong)rec->r.sdir.keyid[1],
+		    rec->r.sdir.pubkey_algo,
+		    (ulong)rec->r.sdir.hintlist );
+	break;
       case RECTYPE_CACH:
 	fprintf(fp, "cach\n");
 	break;
@@ -464,25 +837,32 @@ tdbio_dump_record( TRUSTREC *rec, FILE *fp  )
 int
 tdbio_read_record( ulong recnum, TRUSTREC *rec, int expected )
 {
-    byte buf[TRUST_RECORD_LEN], *p;
+    byte readbuf[TRUST_RECORD_LEN];
+    const byte *buf, *p;
     int rc = 0;
     int n, i;
 
     if( db_fd == -1 )
 	open_db();
-    if( lseek( db_fd, recnum * TRUST_RECORD_LEN, SEEK_SET ) == -1 ) {
-	log_error(_("trustdb: lseek failed: %s\n"), strerror(errno) );
-	return G10ERR_READ_FILE;
-    }
-    n = read( db_fd, buf, TRUST_RECORD_LEN);
-    if( !n ) {
-	return -1; /* eof */
-    }
-    else if( n != TRUST_RECORD_LEN ) {
-	log_error(_("trustdb: read failed (n=%d): %s\n"), n, strerror(errno) );
-	return G10ERR_READ_FILE;
+    buf = get_record_from_cache( recnum );
+    if( !buf ) {
+	if( lseek( db_fd, recnum * TRUST_RECORD_LEN, SEEK_SET ) == -1 ) {
+	    log_error(_("trustdb: lseek failed: %s\n"), strerror(errno) );
+	    return G10ERR_READ_FILE;
+	}
+	n = read( db_fd, readbuf, TRUST_RECORD_LEN);
+	if( !n ) {
+	    return -1; /* eof */
+	}
+	else if( n != TRUST_RECORD_LEN ) {
+	    log_error(_("trustdb: read failed (n=%d): %s\n"), n,
+							strerror(errno) );
+	    return G10ERR_READ_FILE;
+	}
+	buf = readbuf;
     }
     rec->recnum = recnum;
+    rec->dirty = 0;
     p = buf;
     rec->rectype = *p++;
     if( expected && rec->rectype != expected ) {
@@ -508,6 +888,7 @@ tdbio_read_record( ulong recnum, TRUSTREC *rec, int expected )
 	rec->r.ver.validated= buftoulong(p); p += 4;
 	rec->r.ver.keyhashtbl=buftoulong(p); p += 4;
 	rec->r.ver.firstfree =buftoulong(p); p += 4;
+	rec->r.ver.sdirhashtbl =buftoulong(p); p += 4;
 	if( recnum ) {
 	    log_error_f( db_name, "version record with recnum %lu\n",
 							     (ulong)recnum );
@@ -568,7 +949,20 @@ tdbio_read_record( ulong recnum, TRUSTREC *rec, int expected )
 	    rec->r.sig.sig[i].flag = *p++;
 	}
 	break;
-      case RECTYPE_CACH:   /* cache record (FIXME)*/
+      case RECTYPE_SDIR:   /* shadow directory record */
+	rec->r.sdir.lid     = buftoulong(p); p += 4;
+	rec->r.sdir.keyid[0]= buftou32(p); p += 4;
+	rec->r.sdir.keyid[1]= buftou32(p); p += 4;
+	rec->r.sdir.pubkey_algo = *p++;
+	p += 3;
+	rec->r.sdir.hintlist = buftoulong(p);
+	if( rec->r.sdir.lid != recnum ) {
+	    log_error_f( db_name, "sdir LID != recnum (%lu,%lu)\n",
+					 rec->r.sdir.lid, (ulong)recnum );
+	    rc = G10ERR_TRUSTDB;
+	}
+	break;
+      case RECTYPE_CACH:   /* cache record */
 	rec->r.cache.lid    = buftoulong(p); p += 4;
 	memcpy(rec->r.cache.blockhash, p, 20); p += 20;
 	rec->r.cache.trustlevel = *p++;
@@ -596,14 +990,13 @@ tdbio_read_record( ulong recnum, TRUSTREC *rec, int expected )
 
 /****************
  * Write the record at RECNUM
- * FIXME: create/update keyhash record.
  */
 int
 tdbio_write_record( TRUSTREC *rec )
 {
     byte buf[TRUST_RECORD_LEN], *p;
     int rc = 0;
-    int i, n;
+    int i;
     ulong recnum = rec->recnum;
 
     if( db_fd == -1 )
@@ -626,6 +1019,7 @@ tdbio_write_record( TRUSTREC *rec )
 	ulongtobuf(p, rec->r.ver.validated); p += 4;
 	ulongtobuf(p, rec->r.ver.keyhashtbl); p += 4;
 	ulongtobuf(p, rec->r.ver.firstfree ); p += 4;
+	ulongtobuf(p, rec->r.ver.sdirhashtbl ); p += 4;
 	break;
 
       case RECTYPE_FREE:
@@ -677,7 +1071,16 @@ tdbio_write_record( TRUSTREC *rec )
 	}
 	break;
 
-      case RECTYPE_CACH:   /* FIXME*/
+      case RECTYPE_SDIR:
+	ulongtobuf( p, rec->r.sdir.lid); p += 4;
+	u32tobuf( p, rec->r.sdir.keyid[0] ); p += 4;
+	u32tobuf( p, rec->r.sdir.keyid[1] ); p += 4;
+	*p++ = rec->r.sdir.pubkey_algo;
+	p += 3;
+	ulongtobuf( p, rec->r.sdir.hintlist );
+	break;
+
+      case RECTYPE_CACH:
 	ulongtobuf(p, rec->r.cache.lid); p += 4;
 	memcpy(p, rec->r.cache.blockhash, 20); p += 20;
 	*p++ = rec->r.cache.trustlevel;
@@ -700,17 +1103,13 @@ tdbio_write_record( TRUSTREC *rec )
 	BUG();
     }
 
-    if( lseek( db_fd, recnum * TRUST_RECORD_LEN, SEEK_SET ) == -1 ) {
-	log_error(_("trustdb: lseek failed: %s\n"), strerror(errno) );
-	return G10ERR_WRITE_FILE;
-    }
-    n = write( db_fd, buf, TRUST_RECORD_LEN);
-    if( n != TRUST_RECORD_LEN ) {
-	log_error(_("trustdb: write failed (n=%d): %s\n"), n, strerror(errno) );
-	return G10ERR_WRITE_FILE;
-    }
-    else if( rec->rectype == RECTYPE_KEY )
+    rc = put_record_into_cache( recnum, buf );
+    if( rc )
+	;
+    if( rec->rectype == RECTYPE_KEY )
 	rc = update_keyhashtbl( rec );
+    else if( rec->rectype == RECTYPE_SDIR )
+	rc = update_sdirhashtbl( rec );
 
     return rc;
 }
@@ -788,7 +1187,21 @@ tdbio_new_recnum()
 	memset( &rec, 0, sizeof rec );
 	rec.rectype = 0; /* unused record */
 	rec.recnum = recnum;
-	rc = tdbio_write_record( &rec );
+	rc = 0;
+	if( lseek( db_fd, recnum * TRUST_RECORD_LEN, SEEK_SET ) == -1 ) {
+	    log_error(_("trustdb rec %lu: lseek failed: %s\n"),
+						recnum, strerror(errno) );
+	    rc = G10ERR_WRITE_FILE;
+	}
+	else {
+	    int n = write( db_fd, &rec, TRUST_RECORD_LEN);
+	    if( n != TRUST_RECORD_LEN ) {
+		log_error(_("trustdb rec %lu: write failed (n=%d): %s\n"),
+						 recnum, n, strerror(errno) );
+		rc = G10ERR_WRITE_FILE;
+	    }
+	}
+
 	if( rc )
 	    log_fatal_f(db_name,_("failed to append a record: %s\n"),
 						g10_errstr(rc));
@@ -805,13 +1218,13 @@ tdbio_new_recnum()
 int
 tdbio_search_dir_bypk( PKT_public_key *pk, TRUSTREC *rec )
 {
-    byte *fingerprint;
+    byte fingerprint[MAX_FINGERPRINT_LEN];
     size_t fingerlen;
     u32 keyid[2];
     int rc;
 
     keyid_from_pk( pk, keyid );
-    fingerprint = fingerprint_from_pk( pk, NULL, &fingerlen );
+    fingerprint_from_pk( pk, fingerprint, &fingerlen );
     rc = tdbio_search_dir_byfpr( fingerprint, fingerlen,
 				 pk->pubkey_algo, rec );
 
@@ -827,185 +1240,74 @@ tdbio_search_dir_bypk( PKT_public_key *pk, TRUSTREC *rec )
 }
 
 
+static int
+cmp_krec_fpr( void *dataptr, const TRUSTREC *rec )
+{
+    const struct cmp_krec_fpr_struct *d = dataptr;
+
+    return rec->rectype == RECTYPE_KEY
+	   && ( !d->pubkey_algo || rec->r.key.pubkey_algo == d->pubkey_algo )
+	   && rec->r.key.fingerprint_len == d->fprlen
+	   && !memcmp( rec->r.key.fingerprint, d->fpr, d->fprlen );
+}
+
 int
 tdbio_search_dir_byfpr( const byte *fingerprint, size_t fingerlen,
 			int pubkey_algo, TRUSTREC *rec )
 {
+    struct cmp_krec_fpr_struct cmpdata;
     ulong recnum;
     int rc;
-    ulong hashrec, item;
-    int msb;
-    int level=0;
 
     assert( fingerlen == 20 || fingerlen == 16 );
 
     /* locate the key using the hash table */
-    hashrec = get_keyhashrec();
-  next_level:
-    msb = fingerprint[level];
-    hashrec += msb / ITEMS_PER_HTBL_RECORD;
-    rc = tdbio_read_record( hashrec, rec, RECTYPE_HTBL );
-    if( rc ) {
-	log_error( db_name, "scan keyhashtbl failed: %s\n", g10_errstr(rc) );
-	return rc;
+    cmpdata.pubkey_algo = pubkey_algo;
+    cmpdata.fpr = fingerprint;
+    cmpdata.fprlen = fingerlen;
+    rc = lookup_hashtable( get_keyhashrec(), fingerprint, fingerlen,
+			   cmp_krec_fpr, &cmpdata, rec );
+    if( !rc ) {
+	recnum = rec->r.key.lid;
+	/* Now read the dir record */
+	rc = tdbio_read_record( recnum, rec, RECTYPE_DIR);
+	if( rc )
+	    log_error_f(db_name, "can't read dirrec %lu: %s\n",
+						    recnum, g10_errstr(rc) );
     }
-
-    item = rec->r.htbl.item[msb % ITEMS_PER_HTBL_RECORD];
-    if( !item )
-	return -1; /* not found */
-
-    rc = tdbio_read_record( item, rec, 0 );
-    if( rc ) {
-	log_error( db_name, "keyhashtbl read failed: %s\n", g10_errstr(rc) );
-	return rc;
-    }
-    if( rec->rectype == RECTYPE_HTBL ) {
-	hashrec = item;
-	level++;
-	if( level >= fingerlen ) {
-	    log_error( db_name, "keyhashtbl has invalid indirections\n");
-	    return G10ERR_TRUSTDB;
-	}
-	goto next_level;
-    }
-    else if( rec->rectype == RECTYPE_HLST ) {
-	for(;;) {
-	    int i;
-
-	    for(i=0; i < ITEMS_PER_HLST_RECORD; i++ ) {
-		if( rec->r.hlst.rnum[i] ) {
-		    TRUSTREC tmp;
-
-		    rc = tdbio_read_record( rec->r.hlst.rnum[i],
-					     &tmp, RECTYPE_KEY );
-		    if( rc ) {
-			log_error( db_name,
-				   "scan keyhashtbl read key failed: %s\n",
-							     g10_errstr(rc) );
-			return rc;
-		    }
-		    if( (!pubkey_algo || tmp.r.key.pubkey_algo == pubkey_algo)
-			&& tmp.r.key.fingerprint_len == fingerlen
-			&& !memcmp(tmp.r.key.fingerprint,
-					    fingerprint, fingerlen) ) {
-			*rec = tmp;
-			goto found;
-		    }
-		}
-	    }
-	    if( rec->r.hlst.next ) {
-		rc = tdbio_read_record( rec->r.hlst.next, rec, RECTYPE_HLST );
-		if( rc ) {
-		    log_error( db_name,
-			       "scan keyhashtbl read hlst failed: %s\n",
-							 g10_errstr(rc) );
-		    return rc;
-		}
-	    }
-	    else
-		return -1; /* not found */
-	}
-      found:
-	;
-    }
-    else if( rec->rectype == RECTYPE_KEY ) {
-	/* must check that it is the requested key */
-	if( (pubkey_algo && rec->r.key.pubkey_algo != pubkey_algo)
-	    || rec->r.key.fingerprint_len != fingerlen
-	    || memcmp(rec->r.key.fingerprint, fingerprint, fingerlen) )
-	    return -1; /* no: not found */
-    }
-    else {
-	log_error( db_name, "keyhashtbl %lu points to an invalid record\n",
-								item);
-	return G10ERR_TRUSTDB;
-    }
-
-    recnum = rec->r.key.lid;
-    /* Now read the dir record */
-    rc = tdbio_read_record( recnum, rec, RECTYPE_DIR);
-    if( rc )
-	log_error_f(db_name, "can't read dirrec %lu: %s\n",
-						recnum, g10_errstr(rc) );
-
     return rc;
 }
 
+
+
 static int
-del_reclist( ulong recno, int type )
+cmp_sdir( void *dataptr, const TRUSTREC *rec )
 {
-    TRUSTREC rec;
-    int rc;
+    const struct cmp_sdir_struct *d = dataptr;
 
-    while( recno ) {
-	rc = tdbio_read_record( recno, &rec, type);
-	if( rc ) {
-	    log_error_f(db_name, "can't read record %lu: %s\n",
-						recno, g10_errstr(rc));
-	    return rc;
-	}
-	switch( type ) {
-	    case RECTYPE_PREF: recno = rec.r.pref.next; break;
-	    case RECTYPE_UID:  recno = rec.r.uid.next;	break;
-	    default: BUG();
-	}
-	rc = tdbio_delete_record( rec.recnum );
-	if( rc ) {
-	    log_error_f(db_name, "can't delete record %lu: %s\n",
-						rec.recnum, g10_errstr(rc));
-	    return rc;
-	}
-    }
-    return 0;
+    return rec->rectype == RECTYPE_SDIR
+	   && ( !d->pubkey_algo || rec->r.sdir.pubkey_algo == d->pubkey_algo )
+	   && rec->r.sdir.keyid[0] == d->keyid[0]
+	   && rec->r.sdir.keyid[1] == d->keyid[1];
 }
 
-/****************
- * Delete the Userid UIDLID from DIRLID
- */
+
 int
-tdbio_delete_uidrec( ulong dirlid, ulong uidlid )
+tdbio_search_sdir( u32 *keyid, int pubkey_algo, TRUSTREC *rec )
 {
-    TRUSTREC dirrec, rec;
-    ulong recno;
+    struct cmp_sdir_struct cmpdata;
     int rc;
+    byte key[8];
 
-    rc = tdbio_read_record( dirlid, &dirrec, RECTYPE_DIR);
-    if( rc ) {
-	log_error_f(db_name, "can't read dirrec %lu: %s\n", dirlid, g10_errstr(rc));
-	return rc;
-    }
-    recno = dirrec.r.dir.uidlist;
-    for( ; recno; recno = rec.r.uid.next ) {
-	rc = tdbio_read_record( recno, &rec, RECTYPE_UID);
-	if( rc ) {
-	    log_error_f(db_name, "can't read uidrec %lu: %s\n",
-						    recno, g10_errstr(rc));
-	    return rc;
-	}
-	if( recno == uidlid ) {
-	    rc = del_reclist( rec.r.uid.prefrec, RECTYPE_PREF );
-	    if( rc )
-		return rc;
-	    rc = del_reclist( rec.r.uid.siglist, RECTYPE_SIG );
-	    if( rc )
-		return rc;
-	    rc = tdbio_delete_record( recno );
-	    if( rc ) {
-		log_error_f(db_name, "can't delete uidrec %lu: %s\n",
-						    recno, g10_errstr(rc));
-		return rc;
-	    }
-	    dirrec.r.dir.uidlist = 0;
-	    rc = tdbio_write_record( &dirrec );
-	    if( rc ) {
-		log_error_f(db_name, "can't update dirrec %lu: %s\n",
-						  dirrec.recnum, g10_errstr(rc));
-		return rc;
-	    }
-	    return 0;
-	}
-    }
-    return -1; /* not found */
+    /* locate the shadow dir record using the hash table */
+    u32tobuf( key   , keyid[0] );
+    u32tobuf( key+4 , keyid[1] );
+    cmpdata.pubkey_algo = pubkey_algo;
+    cmpdata.keyid[0] = keyid[0];
+    cmpdata.keyid[1] = keyid[1];
+    rc = lookup_hashtable( get_sdirhashrec(), key, 8,
+			   cmp_sdir, &cmpdata, rec );
+    return rc;
 }
 
 
